// fluid simulation example
#include <sys/types.h>
#include <cmath>
#include <vector>
#include <stdlib.h>
#include "Vec3.hpp"

////////////////////
// constants

// particle smoothing radius
const float radius = 0.01;

// size of simulation domain
const Vec3 domain = vec3(1.f, 1.f, 1.f);

// rest density and stiffness
const float baseDensity = 0.01f;
const float stiffness = 1.f;

// gravity strength and direction
const Vec3 gravity = vec3(0.f, 0.f, 1.f);

// number of particles to simulate
const uint32_t movingParticles = 1000;


////////////////////////////////////////
// class definitions

// one particle with unit mass
class Particle {
public:
	// create particle at a position
	Particle(const Vec3 &pos);

	Vec3 position;				  // particle position
	Vec3 velocity;				  // particle velocity
	Vec3 force;					  // forces applied to particle
	float density;				  // local density
	float pressure;				  // local pressure
};

// grid of particle IDs used to find neighbors
class Grid {
public:
	// uniform grid data structure given size of simulation domain and particle radius
	Grid();
	~Grid();

	// give index into pid array for a particle
	uint32_t index(const Vec3 &pos);

	// give cell index in pid array given particle location
	uint32_t gridX, gridY, gridZ, gridSize;	// grid dimensions
	uint32_t strideX, strideY, strideZ; // change to index for change in x/y/z

	// particle IDs within each grid cell
	// id[cell][particle_in_cell] = particle ID
	// id[cell] as fixed allocated array, particle in cell as std::vector
	std::vector<uint32_t> *id;
};

class Particles {
public:
	// initialize particle positions
	Particles();

	// put particles into grid to find neighbors
	void findNeighbors();
	
	// compute local density for each particle
	void computeDensity();

	// compute local pressure for each particle
	void computePressure();

	// force due to gravity and pressure differences
	void computeForces();

	// update position and velocity for timestep of length dt
	void updatePose(float dt);

	// simulate one timestep of length dt
	void simulate(float dt);

	// number of particles
	int size() { return particle.size(); }

private:
	// all of the particles in the system.
	std::vector<Particle> particle;

	// uniform grid of particles, used to find neighbors
	Grid grid;
};

// create particle at a position with zero velocity
Particle::Particle(const Vec3 &pos) : position(pos), velocity(vec3(0.f,0.f,0.f)) {}

// smoothing weight between point p1 and point p2
float weight(const Vec3 &p1, const Vec3 &p2) {
	float dist = length(p2-p1);				// distance from p1 to p2
	float t = fmin(0.f, 1.f - dist/radius); // distance scaled by smoothing radius
	return t*t*t;				// smoothing kernel = (1-dist/radius)^3
}

// gradient of smoothing weight between points p1 and p2
// gradient points in the direction from p1 to p2
Vec3 grad(const Vec3 &p1, const Vec3 &p2) {
	Vec3 dir = p2 - p1;						// direction from p1 to p2
	float dist = length(dir);				// distance between p1 and p2
	float t = fmin(0.f, 1.f - dist/radius); // distance scaled by smoothing radius

	return (-3*t*t) * normalize(dir); // gradient of kernel = -3*(1-dist/radius)^2 in radial direction
}

// allocate uniform grid data structure given size of simulation domain and particle radius
Grid::Grid()
	: gridX(ceil(domain.x/radius) + 2), // with border cells to make boundaries easier
	  gridY(ceil(domain.y/radius) + 2),
	  gridZ(ceil(domain.z/radius) + 2),
	  gridSize(gridX*gridY*gridZ),
	  strideX(gridY*gridZ), strideY(gridZ), strideZ(1)
{
	// auto-growth of std::vector does not work well nested
	// allocate grid once up front
	id = new std::vector<uint32_t>[gridSize];
}

// since grid was manually allocated, be sure to delete it
Grid::~Grid() {
	delete[] id;
}

// give index into pid array for a particle
uint32_t Grid::index(const Vec3 &pos) {
	// dice into grid, with one-cell border on each side
	Vec3 gridPos = pos/radius + 1.f;
	int x = int(gridPos.x), y = int(gridPos.y), z = int(gridPos.z);
	
	// don't return index in the border region
	// this is clamp(x, 1, gridX-2), etc.
	x = x<1 ? 1 : x>gridX-2 ? gridX-2 : x;
	y = y<1 ? 1 : y>gridY-2 ? gridY-2 : y;
	z = z<1 ? 1 : z>gridZ-2 ? gridZ-2 : z;

	// turn into cell index
	return x * strideX + y * strideY + z * strideZ;
}

// initialize particle positions
Particles::Particles() {
	// moving particles in random locations
	for(int i=0; i != movingParticles; ++i)
		particle.push_back(Particle(vec3(drand48() * domain.x, 
										 drand48() * domain.y, 
										 drand48() * domain.z))); 

	// packed grid of non-moving wall particles on x/y walls
	for(float x=0; x < domain.x; x += 0.5*baseDensity) {
		for(float y=0; y < domain.y; y += 0.5*baseDensity) {
			particle.push_back(Particle(vec3(x, y, 0))); 
			particle.push_back(Particle(vec3(x, y, domain.z))); 
		}
	}

	// packed grid of non-moving wall particles on y/z walls
	for(float y=0; y < domain.y; y += 0.5*baseDensity) {
		for(float z=0; z < domain.z; z += 0.5*baseDensity) {
			particle.push_back(Particle(vec3(0, y, z))); 
			particle.push_back(Particle(vec3(domain.x, y, z))); 
		}
	}

	// packed grid of non-moving wall particles on x/z walls
	for(float x=0; x < domain.x; x += 0.5*baseDensity) {
		for(float z=0; z < domain.z; z += 0.5*baseDensity) {
			particle.push_back(Particle(vec3(x, 0, z))); 
			particle.push_back(Particle(vec3(x, domain.y, z))); 
		}
	}
}

// put particles into grid to find neighbors
void Particles::findNeighbors() {
	// clear old grid data
	for(int i=0; i != grid.gridSize; ++i)
		grid.id[i].clear();

	// add each particle to its grid cell
	for(int i=0; i != particle.size(); ++i)
		grid.id[grid.index(particle[i].position)].push_back(i);
}

// compute density of particles
void Particles::computeDensity() {
	// for each particle
	for(int p=0; p != particle.size(); ++p) {
		// where is this particle in the grid?
		uint32_t gridIndex = grid.index(particle[p].position);

		// start with a density of 0
		particle[p].density = 0.f;

		// for each neighboring grid cell
		for(int x=-1; x <= 1; ++x) {
			for(int y=-1; y <= 1; ++y) {
				for(int z=-1; z <= 1; ++z) {
					uint32_t neighborCell = gridIndex + x*grid.strideX + y*grid.strideY + z*grid.strideZ;

					// for each particle in that cell
					for(int cellIdx=0; cellIdx != grid.id[neighborCell].size(); ++cellIdx) {
						uint32_t n = grid.id[neighborCell][cellIdx];

						// add local contribution to density
						// this would be weighted by mass, but I simplified to mass=1
						particle[p].density += weight(particle[n].position,
													  particle[p].position);
					}
				}
			}
		}

		// compute pressure from density
		particle[p].pressure = stiffness * (particle[p].density - baseDensity);
	}
}


// forces due to gravity and pressure differences
void Particles::computeForces() {
	// only for moving particles
	for(int p=0; p != movingParticles; ++p) {
		particle[p].force = gravity;
		
		// where is this particle in the grid?
		uint32_t gridIndex = grid.index(particle[p].position);

		// pressure force from neighbors
		for(int x=-1; x <= 1; ++x) {
			for(int y=-1; y <= 1; ++y) {
				for(int z=-1; z <= 1; ++z) {
					uint32_t neighborCell = gridIndex + x*grid.strideX + y*grid.strideY + z*grid.strideZ;

					// for each particle in that cell
					for(int cellIdx=0; cellIdx != grid.id[neighborCell].size(); ++cellIdx) {
						uint32_t n = grid.id[neighborCell][cellIdx];

						// add local pressure contribution to force
						particle[p].force = particle[p].force +
							grad(particle[n].position, particle[p].position) *
							particle[n].pressure / particle[n].density;
					}
				}
			}
		}
	}
}

// update position and velocity
void Particles::updatePose(float dt) {
	// only for moving particles
	for(int i=0; i != movingParticles; ++i) {
		particle[i].velocity = particle[i].velocity + particle[i].force * dt;
		particle[i].position = particle[i].position + particle[i].velocity * dt;
	}
}

// simulate one timestep of length dt
void Particles::simulate(float dt) {
	// update neighbor lists
	findNeighbors();

	// find local density
	computeDensity();

	// forces due to gravity and pressure differences
	computeForces();

	// update position and velocity
	updatePose(dt);
}


int main() {
	// simulate 100 steps
	Particles p;
	printf("fluid simulation with %d particles\n", p.size());

	for(int step=0; step != 20; ++step) {
		printf("step %d\n", step);
		p.simulate(0.01);
	}

	return 0;
}
